/*
 * QEstEidCommon
 *
 * Copyright (C) 2011-2012 Jargo Kõster <jargo@innovaatik.ee>
 * Copyright (C) 2011-2012 Raul Metsma <raul@innovaatik.ee>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "FileDialog.h"

#include "Common.h"

#include <QtCore/QTemporaryFile>
#include <QtGui/QDesktopServices>
#ifdef Q_OS_WIN
#include <QLibrary>
#include <Shobjidl.h>
#include <Shlguid.h>
#endif

FileDialog::FileDialog( QWidget *parent )
:	QFileDialog( parent )
{
}

FileDialog::Options FileDialog::addOptions()
{
	if( qApp->arguments().contains( "-noNativeFileDialog" ) )
		return DontUseNativeDialog;
	return 0;
}

bool FileDialog::folderIsWritable( const QString &folder )
{
#ifdef Q_OS_WIN
	return QTemporaryFile( folder + "/.XXXXXX" ).open();
#else
	return QFileInfo( folder ).isWritable();
#endif
}

bool FileDialog::fileIsWritable( const QString &filename )
{
	QFile f( filename );
	bool remove = !f.exists();
	bool result = f.open( QFile::WriteOnly|QFile::Append );
	if( remove )
		f.remove();
	return result;
}

QString FileDialog::getDir( const QString &dir )
{
	Q_UNUSED(dir)
#ifndef APPSTORE
	if( !dir.isEmpty() )
		return dir;
	Common *a = qobject_cast<Common*>(QCoreApplication::instance());
	if( a && !a->lastPath().isEmpty() )
		return a->lastPath();
#endif
	return QDesktopServices::storageLocation( QDesktopServices::DocumentsLocation );
}

QString FileDialog::getOpenFileName( QWidget *parent, const QString &caption,
	const QString &dir, const QString &filter, QString *selectedFilter, Options options )
{
	return result( QFileDialog::getOpenFileName( parent,
		caption, getDir( dir ), filter, selectedFilter, options|addOptions() ) );
}

QStringList FileDialog::getOpenFileNames( QWidget *parent, const QString &caption,
	const QString &dir, const QString &filter, QString *selectedFilter, Options options )
{
	return result( QFileDialog::getOpenFileNames( parent,
		caption, getDir( dir ), filter, selectedFilter,
#ifdef Q_OS_WIN
		DontResolveSymlinks|options|addOptions() ) );
#else
		options|addOptions() ) );
#endif
}

QString FileDialog::getExistingDirectory( QWidget *parent, const QString &caption,
	const QString &dir, Options options )
{
#ifdef Q_OS_WIN
	if( QSysInfo::windowsVersion() >= QSysInfo::WV_VISTA )
	{
		QString res;
		IFileOpenDialog *pfd = 0;
		CoCreateInstance( CLSID_FileOpenDialog, NULL, CLSCTX_INPROC, IID_PPV_ARGS(&pfd) );
		pfd->SetOptions( FOS_PICKFOLDERS );
		QString dest = QDir::toNativeSeparators( getDir( dir ) );
		if( !dest.isEmpty() )
		{
			typedef HRESULT (WINAPI *PtrParsingName)(PCWSTR pszPath, IBindCtx *pbc, REFIID riid, void **ppv);
			QLibrary lib("shell32");
			IShellItem *folder = 0;
			PtrParsingName p = PtrParsingName(lib.resolve("SHCreateItemFromParsingName"));
			if( p && SUCCEEDED(p( LPCWSTR(dest.utf16()), 0, IID_PPV_ARGS(&folder) )) )
			{
				pfd->SetFolder(folder);
				folder->Release();
			}
		}
		if( !caption.isEmpty() )
			pfd->SetTitle( LPCWSTR(caption.utf16()) );
		if( SUCCEEDED(pfd->Show( parent && parent->window() ? parent->window()->winId() : 0 )) )
		{
			IShellItem *item = 0;
			if( SUCCEEDED(pfd->GetResult( &item )) )
			{
				LPWSTR path = 0;
				if( SUCCEEDED(item->GetDisplayName( SIGDN_FILESYSPATH, &path )) )
				{
					res = QString( (QChar*)path );
					CoTaskMemFree( path );
				}
				else
				{
					// Case it is Libraries/Favorites
					IEnumShellItems *items = 0;
					if( SUCCEEDED(item->BindToHandler( 0, BHID_EnumItems, IID_PPV_ARGS(&items) )) )
					{
						IShellItem *list = 0;
						if( items->Next( 1, &list, 0 ) == NOERROR )
						{
							LPWSTR path = 0;
							if( SUCCEEDED(list->GetDisplayName( SIGDN_FILESYSPATH, &path )) )
							{
								res = QFileInfo( QString( (QChar*)path ) + "/.." ).absoluteFilePath();
								CoTaskMemFree( path );
							}
							list->Release();
						}
						items->Release();
					}
				}
			}
			item->Release();
		}
		pfd->Release();
		return res;
	}
#endif
	return result( QFileDialog::getExistingDirectory( parent,
		caption, getDir( dir ), ShowDirsOnly|options|addOptions() ) );
}

QString FileDialog::getSaveFileName( QWidget *parent, const QString &caption,
	const QString &dir, const QString &filter, QString *selectedFilter, Options options )
{
#ifdef APPSTORE
	return result( QFileDialog::getSaveFileName( parent,
		caption, QFileInfo( dir ).fileName(), filter, selectedFilter, options|addOptions() ) );
#else
	return result( QFileDialog::getSaveFileName( parent,
		caption, getDir( dir ), filter, selectedFilter, options|addOptions() ) );
#endif
}

QString FileDialog::result( const QString &str )
{
#if defined(Q_OS_MAC)
	QString r = str.normalized( QString::NormalizationForm_C );
#else
	QString r = str;
#endif
	Common *a = qobject_cast<Common*>(QCoreApplication::instance());
	if( a && !r.isEmpty() )
		a->setLastPath( QFileInfo( r ).absolutePath() );
	return r;
}

QStringList FileDialog::result( const QStringList &list )
{
	QStringList l;
	Q_FOREACH( const QString &str, list )
		l << result( str );
	return l;
}
